<?php

/*****************************************************************************

    Radio-4-Matic v0.3 - time-travelling radio system for Raspberry Pi.

     Read more at: https://hylobatidae.org/?action=articleinfo&id=52

------------------------------------------------------------------------------

Copyright (C) 2012-2013 Adam Foster - afoster@hylobatidae.org

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*****************************************************************************/

define( "BUFFER_FILENAME_TEMPLATE", "./buffer.%d.aac" );

define( "SSH_ENABLED", false );

define( "SSH_COMMAND", "" ); // ssh tunnel command goes here
define( "GET_IPLAYER_COMMAND", "get_iplayer --liveradiomode flashaaclow --pid radio:bbc_radio_fourfm --stream --rtmp-liveradio-opts --live" );
define( "FFMPEG_COMMAND", "ffmpeg -f flv -i pipe:0 -acodec copy -vn -f adts -strict -2 -" );
define( "MPLAYER_COMMAND", "mplayer -novideo -nocache -ao alsa " );

// log-files for utilities' stderrs - do a 'tail -f' on a file for magical real-time updates
define( "SSH_LOG", "ssh.log" );
define( "GET_IPLAYER_LOG", "get_iplayer.log" );
define( "FFMPEG_LOG", "ffmpeg.log" );
define( "MPLAYER_LOG", "mplayer.log" );

define( "RECORDING_STAGE_WAIT", 0 );
define( "RECORDING_STAGE_START_CAPTURE", 1 );
define( "RECORDING_STAGE_RUNNING", 2 );
define( "RECORDING_STAGE_SHUTDOWN", 3 );

// mystery streaming-delay removal - gets the pips a bit closer to where they ought to be, if needed
define( "RECORDING_STREAMING_DELAY", 0 );

define( "PLAYBACK_STAGE_WAIT", 0 );
define( "PLAYBACK_STAGE_START_MPLAYER", 1 );
define( "PLAYBACK_STAGE_RUNNING", 2 );
define( "PLAYBACK_STAGE_SHUTDOWN", 3 );

define( "RECORDING_TIMEZONE", "Europe/London" );
define( "PLAYBACK_TIMEZONE", "America/Los_Angeles" );

// how many seconds to wait on each iteration of main loop (when data is arriving, it goes round every update)
// - this is mainly to stop it zooming round as quickly as possibly, saturating the CPU
define( "LOOP_DELAY", 1 );

// how long to wait for data from get_iplayer before restarting - rtmpdump has a tendency to get stuck
define( "PROCESS_TIMEOUT", 30 );

// process information gets shoved in these
$ssh = null;
$get_iplayer = null;
$ffmpeg = null;
$mplayer = null;

// initial stages for big select() whatsits
$recording_stage = RECORDING_STAGE_WAIT;
$playback_stage = PLAYBACK_STAGE_WAIT;

$buffer_list = array();

// time-ranges to record - e.g. 6am to 12pm, 6pm to 1am
$recording_time_range_list = array(
	array( 'start'=>6,  'end'=>12 ),
	array( 'start'=>18, 'end'=>1 )
);

internal_log( "INIT", "Radio-4-Matic v0.3 - afoster@hylobatidae.org" );

while ( true )
{
	recording( $recording_stage );
	playback( $playback_stage );
}

// ===========================================================================
	
function recording( &$recording_stage )
{
	global $ssh;
	global $get_iplayer;
	global $ffmpeg;
	global $mplayer;
	
	global $buffer_list;
	
	global $recording_time_range_list;
	
	$recording_active_flag = false;
	
	date_default_timezone_set( RECORDING_TIMEZONE ); // need to find out the current hour in Britain-land
	
	$hour = (integer )date( 'G' );
	
	foreach ( $recording_time_range_list as $range )
	{
		if ( $range['start'] > $range['end'] ) // so we can record from 1800h to 0100h, for example
		{
			if ( $hour >= $range['start'] || $hour < $range['end'] )
				$recording_active_flag = true;
		}
		else
		{
			if ( $hour >= $range['start'] && $hour < $range['end'] )
				$recording_active_flag = true;
		}
	}
	
	date_default_timezone_set( PLAYBACK_TIMEZONE ); // switch back to Seattle-land
	
	switch ( $recording_stage )
	{
		case RECORDING_STAGE_WAIT:
		
			if ( $recording_active_flag )
			{
				internal_log( "REC", "In time range - starting recording!" );
				$recording_stage = RECORDING_STAGE_START_CAPTURE;
			}
			else
				sleep( LOOP_DELAY ); // don't want to go round the main loop too quickly
				
			break;
		
		case RECORDING_STAGE_START_CAPTURE:
		
			if ( SSH_ENABLED )
			{
				internal_log( "REC", "Starting ssh tunnel." );
			
				if ( file_exists( SSH_LOG ) )
					external_log_rotate( SSH_LOG );
					
				$ssh = start(
					SSH_COMMAND,
					array(
						0=>array( "pipe", 'r' ), //stdin
						1=>array( "file", "/dev/null", 'a' ), //stdout
						2=>array( "file", SSH_LOG, 'a' ) //stderr
					)
				);
			}
							
			internal_log( "REC", "Starting get_iplayer." );
			
			if ( file_exists( GET_IPLAYER_LOG ) )
				external_log_rotate( GET_IPLAYER_LOG );
		
			$get_iplayer = start(
				GET_IPLAYER_COMMAND,
				array(
					0=>array( "pipe", 'r' ),
					1=>array( "pipe", 'w' ),
					2=>array( "file", GET_IPLAYER_LOG, 'a' )
				)
			);
				
			stream_set_blocking( $get_iplayer['pipes'][0], 0 ); // non-blocking IO - freads from iplayer can't block other tasks
			stream_set_blocking( $get_iplayer['pipes'][1], 0 );
			
			$recording_stage = RECORDING_STAGE_RUNNING;
			
			break;
			
		case RECORDING_STAGE_RUNNING:
			
			$select_read = array( $get_iplayer['pipes'][1] );
			$select_write = NULL;
			$select_except = NULL;
			
			stream_select( $select_read, $select_write, $select_except, LOOP_DELAY ); // wait for data, or timeout
			
			$audio_data = fread( $get_iplayer['pipes'][1], 8192 ); // read up to 8kB from iplayer's stdout
			
			$audio_bytes_total = strlen( $audio_data );
			$audio_bytes_remaining = $audio_bytes_total;
			
			if ( $audio_bytes_total > 0 ) // has data arrived? (may have just timed out instead)
			{
				update_response( $get_iplayer ); // it has! update iplayer response timestamp
				
				if ( !is_started ( $ffmpeg ) ) // if we have data and ffmpeg hasn't been started yet, start it
				{
					internal_log( "REC", "get_iplayer has started returning data." );
					internal_log( "REC", "Starting ffmpeg." );
					
					if ( file_exists( FFMPEG_LOG ) )
						external_log_rotate( FFMPEG_LOG );
					
					// note magical fudge-factor to reduce the natural delay from streaming live radio

					$buffer_start_time = time() + RECORDING_STREAMING_DELAY;
					$buffer_filename = sprintf( BUFFER_FILENAME_TEMPLATE, $buffer_start_time );
					
					internal_log( "REC", "Creating new buffer '".$buffer_filename."'." );
					
					// stick timestamp and filename on to end of list
					array_push( $buffer_list, array( 'timestamp'=>$buffer_start_time, 'filename'=>$buffer_filename ) );
					
					$ffmpeg = start(
						FFMPEG_COMMAND,
						array(
							0=>array( "pipe", 'r' ),	// stdin
							1=>array( "file", $buffer_filename, 'w' ),	// stdout
							2=>array( "file", FFMPEG_LOG, 'a' ) // stderr
						)
					);
						
					stream_set_blocking( $ffmpeg['pipes'][0], 0 );
				}
				
				// non-blocking IO on ffmpeg's stdin means it doesn't necessarily take all of it in one go
				// - keep stuffing data in until it's all gone
				
				while ( is_running( $ffmpeg ) && ( $audio_bytes_remaining > 0 ) ) 				{
					$audio_bytes_written = fwrite( $ffmpeg['pipes'][0], substr( $audio_data, $audio_bytes_total - $audio_bytes_remaining, $audio_bytes_remaining ) );
					
					$audio_bytes_remaining -= $audio_bytes_written;
				}
			}
			
			// housekeeping to make sure everything is still tickety-boo
			
			if ( SSH_ENABLED && !is_running( $ssh ) )
			{
				internal_log( "REC", "ssh appears to have died." );
				$recording_stage = RECORDING_STAGE_SHUTDOWN;
			}
			
			if ( !is_responding( $get_iplayer ) )
			{
				internal_log( "REC", "get_iplayer hasn't returned data in a while." );
				$recording_stage = RECORDING_STAGE_SHUTDOWN;
			}
			
			if ( !is_running( $get_iplayer ) )
			{
				internal_log( "REC", "get_iplayer appears to have died." );
				$recording_stage = RECORDING_STAGE_SHUTDOWN;
			}
			
			if ( is_started( $ffmpeg ) && !is_running( $ffmpeg ) )
			{
				internal_log( "REC", "ffmpeg appears to have died." );
				$recording_stage = RECORDING_STAGE_SHUTDOWN;
			}
			
			if ( !$recording_active_flag )
			{
				internal_log( "REC", "Outside time range - stopping recording!" );
				$recording_stage = RECORDING_STAGE_SHUTDOWN;
			}
			
			break;
			
		case RECORDING_STAGE_SHUTDOWN: // shut. down. everything.
				
			if ( is_started( $ffmpeg ) )
			{
				internal_log( "REC", "Stopping ffmpeg." );
				stop( $ffmpeg );
			}
				
			if ( is_started( $get_iplayer ) )
			{
				internal_log( "REC", "Stopping get_iplayer." );
				stop( $get_iplayer );
			}
			
			// ssh takes a remarkably long time to timeout if network stuff resets - restart it anyway if anything's died
			// (get_iplayer might have exited due to a transient network problem, which ssh won't notice for a while)
			if ( SSH_ENABLED && is_started( $ssh ) ) 
			{
				internal_log( "REC", "Stopping ssh." );
				stop( $ssh );
			}
			
			$recording_stage = RECORDING_STAGE_WAIT;
			
			sleep( LOOP_DELAY ); // don't want to go round the main loop too quickly
				
			break;
			
		default:
		
			die( "recording: How did I get here?\n" );
	}
	
	return; // back to main loop...
}

// ===========================================================================

function playback( &$playback_stage )
{
	global $mplayer;
	
	global $buffer_list;
	
	$current_buffer = reset( $buffer_list ); // get first item in list (lol PHP)
	
	$playback_delay = get_timezone_offset( PLAYBACK_TIMEZONE, RECORDING_TIMEZONE ); // seconds between London and Seattle
	
	if ( $playback_delay < 0 )
		die( "playback: delay is negative - can't travel back in time (yet)!\n" );
	
	switch ( $playback_stage )
	{
		case PLAYBACK_STAGE_WAIT:
		
			// is there a buffer waiting to play? if so, wait until it's time to play it
			if ( is_array( $current_buffer ) && ( $current_buffer['timestamp'] + $playback_delay <= time() ) )
			{
				$playback_stage = PLAYBACK_STAGE_START_MPLAYER;
			}
			
			break;
			
		case PLAYBACK_STAGE_START_MPLAYER:
		
			if ( is_started( $mplayer ) ) // shouldn't happen, but just in case...
			{
				internal_log( "PLAY", "mplayer already running - closing." );
				stop( $mplayer );
			}
			
			internal_log( "PLAY", "Started mplayer on buffer '".$current_buffer['filename']."'." );
			
			if ( file_exists( MPLAYER_LOG ) )
				external_log_rotate( MPLAYER_LOG );
				
			$mplayer = start(
				MPLAYER_COMMAND.escapeshellarg( $current_buffer['filename'] ),
				array( 
					0=>array( "pipe", 'r' ), //stdin
					1=>array( "file", "/dev/null", 'a' ), //stdout
					2=>array( "file", MPLAYER_LOG, 'a' ) //stderr
				)
			);
			
			$playback_stage = PLAYBACK_STAGE_RUNNING;
			
			break;
			
		case PLAYBACK_STAGE_RUNNING:
		
			// basically do nothing except check if mplayer is done
			
			if ( !is_running( $mplayer ) ) // if it's exited, either the buffer has finished playing or the buffer was bad
				$playback_stage = PLAYBACK_STAGE_SHUTDOWN;
				
			break;
			
		case PLAYBACK_STAGE_SHUTDOWN:
			
			internal_log( "PLAY", "Stopping mplayer." );
			stop( $mplayer );
			
			internal_log( "PLAY", "Deleting obsolete buffer '".$current_buffer['filename']."'." );
			
			unlink( $current_buffer['filename'] );
			array_shift( $buffer_list ); // remove first item from list
			
			$playback_stage = PLAYBACK_STAGE_WAIT;
			
			break;
			
		default:
		
			die( "playback: How did I get here?\n" );
	}
	
	return; // back to main loop...
}

// ===========================================================================

function start( $command, $descriptor_spec )
{
	// cheesy wrapper for proc_open, stuffing results into array with a few extra goodies
	
	$info = array();
	
	$info['process'] = proc_open( $command, $descriptor_spec, $info['pipes'] );	
	$info['started'] = true;
	
	update_response( $info );
	
	return ( $info );
}
	
// ===========================================================================

function is_started( $info )
{
	// have we started this process? (conceptually different from 'is it running' - we may have started it but it's since died)
	
	if ( isset( $info['started'] ) && $info['started'] )
		return ( true );
	else
		return ( false );
}

// ===========================================================================

function is_running( $info )
{
	if ( !is_started( $info ) )
		return ( false );
		
	$status = proc_get_status( $info['process'] );
	
	if ( $status['running'] )
		return ( true );
	else
		return ( false );
}

// ===========================================================================

function update_response( &$info )
{
	// update response time for is_responding()
	
	$info['last_response'] = time();
	
	return;
}

// ===========================================================================

function is_responding( $info )
{
	// cheesy hack for finding out when get_iplayer last returned data - rtmpdump would often get stuck, remaining alive but
	// not doing anything
	
	if ( time() > $info['last_response'] + PROCESS_TIMEOUT )
		return ( false );
	else
		return ( true );
}

// ===========================================================================

function stop( &$info )
{
	$status = proc_get_status( $info['process'] );

	foreach ( $info['pipes'] as $pipe ) // close any opened pipes
	{
		fclose( $pipe );
	}
	
	posix_kill( $status['pid'], 9 ); // make sure process is very dead, soon (get_iplayer often wouldn't respond to a usual close)
	proc_close( $info['process'] );
	
	$info = null;
}


// ===========================================================================
// function by '[d][a][n][at][authenticdesign][.][net]' at
// http://php.net/manual/en/function.timezone-offset-get.php

function get_timezone_offset( $remote_tz, $origin_tz = null )
{
    if ( $origin_tz === null )
    {
        if( !is_string( $origin_tz = date_default_timezone_get() ) )
        {
            return false; // A UTC timestamp was returned -- bail out!
        }
    }
    
    $origin_dtz = new DateTimeZone( $origin_tz );
    $remote_dtz = new DateTimeZone( $remote_tz );
    $origin_dt = new DateTime( "now", $origin_dtz );
    $remote_dt = new DateTime( "now", $remote_dtz );
    $offset = $origin_dtz->getOffset( $origin_dt ) - $remote_dtz->getOffset( $remote_dt );
    
    return ( $offset );
}

// ===========================================================================

function internal_log( $module, $message )
{
	date_default_timezone_set( PLAYBACK_TIMEZONE ); 

	print( " [".$module."]\t".date( 'c' )." - ".$message."\n" );
	
	return;
}

// ===========================================================================

function external_log_rotate( $filename )
{
	// whatever.log.[n] renamed to whatever.log.[n+1], in reverse order so only overwrite oldest logfile
	for ( $n = 8; $n >= 1; $n-- )
	{
		if ( file_exists( $filename.".".$n ) )
			rename( $filename.".".$n, $filename.".".( $n + 1 ) );
	}
	
	// whatever.log to whatever.log.1
	
	rename( $filename, $filename.".1" );
}

?>